use itertools::Itertools;
use rand::prelude::*;
use rand::rngs::StdRng;
use std::collections::HashSet;
use std::io::Write;

/// un graphe tout simple.
/// les sommets sont des usize de 0 a `aretes.len() -1`.
struct Graphe {
    /// pour chaque sommet un vecteur de tous les sommets voisins
    aretes: Vec<Vec<usize>>,
}

impl Graphe {
    /// genere un graphe aleatoire
    fn aleatoire<R: Rng>(generateur: &mut R, nombre_sommets: usize) -> Self {
        Graphe {
            aretes: std::iter::repeat_with(|| {
                std::iter::repeat_with(|| generateur.gen::<usize>() % nombre_sommets)
                    .unique()
                    .take(2)
                    .collect()
            })
            .take(nombre_sommets)
            .collect(),
        }
    }
    /// iterateur sur tous les sommets
    fn sommets(&self) -> impl Iterator<Item = usize> {
        0..self.aretes.len()
    }
    /// iterateur sur tous les voisins directs d'un sommet donne
    fn voisins<'a>(&'a self, s: usize) -> impl Iterator<Item = usize> + 'a {
        self.aretes[s].iter().copied()
    }
    /// iterateur sur tous les sommets accessibles à partir d'un point de départ.
    /// on ne passe pas deux fois sur le meme sommet.
    fn sommets_accessibles(&self, depart: usize) -> Accessibles {
        Accessibles {
            reste: vec![depart],
            graphe: self,
            vus: Default::default(), // on se rappelle la ou on est deja passe
        }
    }
    /// affiche le graphe dans firefox a l'aide de graphviz
    fn affichage(&self) -> std::io::Result<()> {
        let id = rand::random::<u32>();
        let mut chemin_dot = std::env::temp_dir();
        chemin_dot.push(id.to_string());
        let mut chemin_png = chemin_dot.clone();
        chemin_dot.set_extension("dot");
        chemin_png.set_extension("png");
        let mut dot = std::fs::File::create(&chemin_dot)?;
        writeln!(&mut dot, "digraph g {{")?;
        self.sommets().try_for_each(|s| {
            self.voisins(s)
                .try_for_each(|v| writeln!(&mut dot, "n{} -> n{};", s, v))
        })?;
        write!(&mut dot, "}}")?;
        std::process::Command::new("dot")
            .arg("-T")
            .arg("png")
            .arg("-o")
            .arg(&chemin_png)
            .arg(&chemin_dot)
            .status()?;
        std::process::Command::new("firefox")
            .arg(&chemin_png)
            .status()?;
        Ok(())
    }
}

/// iterateur profondeur d'abord sur tous les sommets accessibles
struct Accessibles<'a> {
    reste: Vec<usize>,
    graphe: &'a Graphe,
    vus: HashSet<usize>,
}

impl<'a> Accessibles<'a> {
    /// pop les sommets du reste, un par un et itere sur tous les sommets non encore vus.
    /// paresseux !
    fn extraire_reste_non_vu<'b>(&'b mut self) -> impl Iterator<Item = usize> + 'b {
        unimplemented!()
    }
}

impl<'a> Iterator for Accessibles<'a> {
    type Item = usize;
    fn next(&mut self) -> Option<Self::Item> {
        let suivant = self.extraire_reste_non_vu().next();
        suivant.map(|s| {
            self.vus.insert(s);
            // note: ca serait mieux de filtrer ici
            // mais l'exo est fait comme ca.
            // on ne changera donc pas ce code
            self.reste.extend(self.graphe.voisins(s));
            s
        })
    }
}

fn main() {
    let mut rng = StdRng::seed_from_u64(0);
    let g = Graphe::aleatoire(&mut rng, 8);
    g.affichage().unwrap_or_else(|_| {
        println!("il faut 'graphviz (dot)' et 'firefox' pour visualiser le graphe")
    });
    assert_eq!(
        g.sommets_accessibles(0).collect::<HashSet<_>>(),
        vec![0, 2, 3, 4, 6].into_iter().collect::<HashSet<_>>()
    );
}
